/*
 Copyright 2020 Masarykova univerzita

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0


Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
 */

package cz.fi.muni.ANALYZA.db_orchestrator_filler.GUI;

import com.github.dockerjava.api.DockerClient;
import com.github.dockerjava.api.model.BuildResponseItem;
import com.github.dockerjava.api.model.PushResponseItem;
import com.github.dockerjava.core.DefaultDockerClientConfig;
import com.github.dockerjava.core.DockerClientBuilder;
import com.github.dockerjava.core.command.BuildImageResultCallback;
import com.github.dockerjava.core.command.PushImageResultCallback;
import javafx.concurrent.Task;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.Collections;
import java.util.Properties;
import java.util.concurrent.TimeUnit;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class BuildingImageTask extends Task<Void> {

    private final File file;
    private final String repositoryName;
    private final String tag;
    private Properties properties;

    public BuildingImageTask(File file, String repositoryName, String tag) {
        this.file = file;
        this.repositoryName = repositoryName;
        this.tag = tag;

        try (InputStream input = BuildingImageTask.class.getClassLoader().getResourceAsStream("config.properties")) {

            properties = new Properties();

            if (input == null) {
                System.out.println("Sorry, unable to find configuration file.");
                return;
            }
            properties.load(input);

        } catch (IOException ex) {
            ex.printStackTrace();
        }
    }

    public String getImageName(){
        return repositoryName + ":" + tag;
    }

    @Override
    protected Void call() throws Exception {
        DefaultDockerClientConfig config = DefaultDockerClientConfig.createDefaultConfigBuilder()
                .withRegistryUrl(properties.getProperty("docker.registry.url"))
                .withRegistryUsername(properties.getProperty("docker.registry.username"))
                .withRegistryPassword(properties.getProperty("docker.registry.password"))
                .withDockerTlsVerify(true)
                .withDockerCertPath(properties.getProperty("docker.certificate.filePath"))
                .withDockerHost(properties.getProperty("docker.host"))
                .build();

        DockerClient dockerClient = DockerClientBuilder.getInstance(config).build();

        BuildImageResultCallback callback = new BuildImageResultCallback() {
            @Override
            public void onNext(BuildResponseItem item) {
                System.out.println("It's done ->" + item);
                String message = item.toString();
                String regex = ".*stream=Step (.*)/(.*) : .*";
                Pattern pattern = Pattern.compile(regex);
                Matcher matcher = pattern.matcher(message);
                while (matcher.find() ) {
                    updateProgress(Integer.parseInt(matcher.group(1)) , Integer.parseInt(matcher.group(2)));
                    System.out.println(matcher.group(1));
                    System.out.println(matcher.group(2));
                }

                if (isCancelled()) {
                    try {
                        dockerClient.close();
                    } catch (IOException e) {
                        e.printStackTrace();
                    }
                    return;
                }
                super.onNext(item);
            }
        };

        PushImageResultCallback pushImageResultCallback = new PushImageResultCallback() {
            @Override
            public void onNext(PushResponseItem item) {
                System.out.println("It's done too - >" + item);
                super.onNext(item);
            }
        };

        dockerClient.buildImageCmd(file)
                .withTags(Collections.singleton("orchestrator:443/" + repositoryName + ":" + tag))
                .exec(callback)
                .awaitImageId();

        System.out.println("\n x \n x \n x \n");

        dockerClient.pushImageCmd("orchestrator:443/" + repositoryName)
                .withTag(tag)
                .exec(pushImageResultCallback)
                .awaitCompletion(120, TimeUnit.SECONDS);
        updateProgress(1,1);

        return null;
    }
}
